*********************************************
* Program to estimate Hamilton's Markov     *
* switching model using the EM algorithm.   *
* 					                             *
* Cite:                                     *
* Bruce Mizrach and James Watkins           *
* "A Markov Switching Cookbook"             *
*                                           *
* in Philip Rothman (ed.)                   *
* "Nonlinear Time Series Analysis of        *
* Economics and Financial Data"             *
* Dordrecht: Kluwer, 1999, 33-43.           *
*********************************************
allocate 135
open data dgdp.wk1
data(org=obs,format=wks) / dgdp
set y = dgdp
********** variable declarations **********
declare vector p(6) lp(6) th(9) lth(9)
declare real p11 p22 v q1 q2
declare rectangular r4(16,135) r5(32,135)
declare rectangular r6(64,135) r7(128,135)
declare rectangular r8(256,135) r9(512,135)
declare rectangular r10(1024,135)
declare rectangular b(32,135)
declare rectangular e(32,135) vd(1024,135)
declare vector ucd(135) c(135)
declare rectangular h(32,135)
declare real sum llval wsr c1 c2
declare vector s11(135) s21(135) s12(135) s22(135)
declare vector s1(135) s2(135)
declare real sums11 sums22 sums1 sums2
*
declare vector ap(6) eps(6)
declare vector dp(6)
declare real g1 g2 s
declare vector g(6)
declare vector gs(10) vals(10)
declare real max
*
declare integer i k m n w j itr sa
declare real dthmax
********** state functions ***********
frml e1 = y(T) - p(1)
frml e2 = y(T) - p(2)
* (01):1,1,1,1,1
frml f01 = (e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (02):2,1,1,1,1
frml f02 = (e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (03):1,2,1,1,1
frml f03 = (e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (04):2,2,1,1,1
frml f04 = (e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (05):1,1,2,1,1
frml f05 = (e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (06):2,1,2,1,1
frml f06 = (e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (07):1,2,2,1,1
frml f07 = (e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (08):2,2,2,1,1
frml f08 = (e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2
* (09):1,1,1,2,1
frml f09 = (e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (10):2,1,1,2,1
frml f10 = (e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (11):1,2,1,2,1
frml f11 = (e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (12):2,2,1,2,1
frml f12 = (e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (13):1,1,2,2,1
frml f13 = (e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (14):2,1,2,2,1
frml f14 = (e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (15):1,2,2,2,1
frml f15 = (e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (16):2,2,2,2,1
frml f16 = (e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2
* (17):1,1,1,1,2
frml f17 = (e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (18):2,1,1,1,2
frml f18 = (e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (19):1,2,1,1,2
frml f19 = (e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (20):2,2,1,1,2
frml f20 = (e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (21):1,1,2,1,2
frml f21 = (e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (22):2,1,2,1,2
frml f22 = (e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (23):1,2,2,1,2
frml f23 = (e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (24):2,2,2,1,2
frml f24 = (e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2
* (25):1,1,1,2,2
frml f25 = (e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (26):2,1,1,2,2
frml f26 = (e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (27):1,2,1,2,2
frml f27 = (e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (28):2,2,1,2,2
frml f28 = (e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (29):1,1,2,2,2
frml f29 = (e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (30):2,1,2,2,2
frml f30 = (e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (31):1,2,2,2,2
frml f31 = (e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
* (32):2,2,2,2,2
frml f32 = (e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2
********** residuals procedure **********
procedure resids
do m = 5,135
     compute e(1,m)  = f01(m) ; compute e(2,m)  = f02(m)
     compute e(3,m)  = f03(m) ; compute e(4,m)  = f04(m)
     compute e(5,m)  = f05(m) ; compute e(6,m)  = f06(m)
     compute e(7,m)  = f07(m) ; compute e(8,m)  = f08(m)
     compute e(9,m)  = f09(m) ; compute e(10,m) = f10(m)
     compute e(11,m) = f11(m) ; compute e(12,m) = f12(m)
     compute e(13,m) = f13(m) ; compute e(14,m) = f14(m)
     compute e(15,m) = f15(m) ; compute e(16,m) = f16(m)
     compute e(17,m) = f17(m) ; compute e(18,m) = f18(m)
     compute e(19,m) = f19(m) ; compute e(20,m) = f20(m)
     compute e(21,m) = f21(m) ; compute e(22,m) = f22(m)
     compute e(23,m) = f23(m) ; compute e(24,m) = f24(m)
     compute e(25,m) = f25(m) ; compute e(26,m) = f26(m)
     compute e(27,m) = f27(m) ; compute e(28,m) = f28(m)
     compute e(29,m) = f29(m) ; compute e(30,m) = f30(m)
     compute e(31,m) = f31(m) ; compute e(32,m) = f32(m)
end do m
end resids
********** distribution procedure **********
procedure dist
compute c1 = 1/((2*%pi*abs(v))**.5)
compute c2 = (-1)/(2*abs(v))
do m = 5,135
     do n = 1,32
          compute vd(n,m) = c1*exp(c2*e(n,m))
     end do n
end do m
do m = 5,135
     do n = 33,1024
          compute vd(n,m) = vd(n-32,m)
     end do n
end do m
end dist
********** rho procedure **********
procedure rho
compute q1 = (1-p22)/(2-p11-p22)
compute q2 = (1-p11)/(2-p11-p22)
*
compute r4(1,4)  = p11*p11*p11*q1
compute r4(2,4)  = (1-p11)*p11*p11*q1
compute r4(3,4)  = (1-p22)*(1-p11)*p11*q1
compute r4(4,4)  = p22*(1-p11)*p11*q1
compute r4(5,4)  = p11*(1-p22)*(1-p11)*q1
compute r4(6,4)  = (1-p11)*(1-p22)*(1-p11)*q1
compute r4(7,4)  = (1-p22)*p22*(1-p11)*q1
compute r4(8,4)  = p22*p22*(1-p11)*q1
*
compute r4(9,4)  = p11*p11*(1-p22)*q2
compute r4(10,4) = (1-p11)*p11*(1-p22)*q2
compute r4(11,4) = (1-p22)*(1-p11)*(1-p22)*q2
compute r4(12,4) = p22*(1-p11)*(1-p22)*q2
compute r4(13,4) = p11*(1-p22)*p22*q2
compute r4(14,4) = (1-p11)*(1-p22)*p22*q2
compute r4(15,4) = (1-p22)*p22*p22*q2
compute r4(16,4) = p22*p22*p22*q2
end rho
********** filter procedure **********
procedure filter
do n = 5,135
     compute sum = 0
     do m = 1,29,4
          compute w = (m+1)/2
          compute b(m,n)   = vd(m,n)*p11*r4(w,n-1)
          compute b(m+1,n) = vd(m+1,n)*(1-p11)*r4(w,n-1)
          compute b(m+2,n) = vd(m+2,n)*(1-p22)*r4(w+1,n-1)
          compute b(m+3,n) = vd(m+3,n)*p22*r4(w+1,n-1)
          compute sum = sum + b(m,n) + b(m+1,n) + b(m+2,n) + b(m+3,n)
     end do m
     compute ucd(n) = sum
     compute c(n) = (ucd(n))**(-1)
     do m = 1,16
          compute r5(m,n)    = b(m,n)*c(n)
          compute r5(m+16,n) = b(m+16,n)*c(n)
          compute r4(m,n) = r5(m,n) + r5(m+16,n)
     end do m
end do n
end filter
********** smoothing procedure **********
procedure smoother
do k = 6,135
     do m = 1,61,4
          compute w = (m+1)/2
          compute r6(m,k)   = vd(m,k)*p11*r5(w,k-1)*c(k)
          compute r6(m+1,k) = vd(m+1,k)*(1-p11)*r5(w,k-1)*c(k)
          compute r6(m+2,k) = vd(m+2,k)*(1-p22)*r5(w+1,k-1)*c(k)
          compute r6(m+3,k) = vd(m+3,k)*p22*r5(w+1,k-1)*c(k)
     end do m
end do k
*
do k = 7,135
     do m = 1,125,4
          compute w = (m+1)/2
          compute r7(m,k)   = vd(m,k)*p11*r6(w,k-1)*c(k)
          compute r7(m+1,k) = vd(m+1,k)*(1-p11)*r6(w,k-1)*c(k)
          compute r7(m+2,k) = vd(m+2,k)*(1-p22)*r6(w+1,k-1)*c(k)
          compute r7(m+3,k) = vd(m+3,k)*p22*r6(w+1,k-1)*c(k)
     end do m
end do k
*
do k = 8,135
     do m = 1,253,4
          compute w = (m+1)/2
          compute r8(m,k)   = vd(m,k)*p11*r7(w,k-1)*c(k)
          compute r8(m+1,k) = vd(m+1,k)*(1-p11)*r7(w,k-1)*c(k)
          compute r8(m+2,k) = vd(m+2,k)*(1-p22)*r7(w+1,k-1)*c(k)
          compute r8(m+3,k) = vd(m+3,k)*p22*r7(w+1,k-1)*c(k)
     end do m
end do k
*
do k = 9,135
     do m = 1,509,4
          compute w = (m+1)/2
          compute r9(m,k)   = vd(m,k)*p11*r8(w,k-1)*c(k)
          compute r9(m+1,k) = vd(m+1,k)*(1-p11)*r8(w,k-1)*c(k)
          compute r9(m+2,k) = vd(m+2,k)*(1-p22)*r8(w+1,k-1)*c(k)
          compute r9(m+3,k) = vd(m+3,k)*p22*r8(w+1,k-1)*c(k)
     end do m
end do k
*
do k = 10,135
     do m = 1,1021,4
          compute w = (m+1)/2
          compute r10(m,k)   = vd(m,k)*p11*r9(w,k-1)*c(k)
          compute r10(m+1,k) = vd(m+1,k)*(1-p11)*r9(w,k-1)*c(k)
          compute r10(m+2,k) = vd(m+2,k)*(1-p22)*r9(w+1,k-1)*c(k)
          compute r10(m+3,k) = vd(m+3,k)*p22*r9(w+1,k-1)*c(k)
     end do m
end do k
*
do j = 5,130
     do k = 1,993,32
          compute sum = 0
          do m = 0,31
               compute sum = sum + r10(k+m,j+5)
          end do m
          compute w = (k+31)/32
          compute h(w,j) = sum
    end do k
end do j
*
do k = 1,32
     compute h(k,135) = r5(k,135)
end do k
*
do k = 1,63,2
     compute w = (k+1)/2
     compute h(w,134) = r6(k,135) + r6(k+1,135)
end do k
*
do k = 1,125,4
     compute sum = 0
     do m = 0,3
          compute sum = sum + r7(k+m,135)
     end do m
     compute w = (k+3)/4
     compute h(w,133) = sum
end do k
*
do k = 1,249,8
     compute sum = 0
     do m = 0,7
          compute sum = sum + r8(k+m,135)
     end do m
     compute w = (k+7)/8
     compute h(w,132) = sum
end do k
*
do k = 1,497,16
     compute sum = 0
     do m = 0,15
          compute sum = sum + r9(k+m,135)
     end do m
     compute w = (k+15)/16
     compute h(w,131) = sum
end do k
end smoother
********** weighted sq. resids procedure **********
procedure pwsr
execute resids
compute sum = 0
do n = 5,135
     do m = 1,32
          compute sum = sum + e(m,n)*h(m,n)
     end do m
end do n
compute wsr = sum*(-1)
end pwsr
********** variance procedure **********
procedure var
execute resids
compute sum = 0
do n = 5,135
     do m = 1,32
          compute sum = sum + e(m,n)*h(m,n)
     end do m
end do n
compute v = sum/131
end var
********** transition probabilities procedure **********
procedure trans
do k = 6,135
     compute sum = 0
     do m = 1,29,4
          compute sum = sum + h(m,k)
     end do m
     compute s11(k) = sum
*
     compute sum = 0
     do m = 2,30,4
          compute sum = sum + h(m,k)
     end do m
     compute s21(k) = sum
*
     compute sum = 0
     do m = 3,31,4
          compute sum = sum + h(m,k)
     end do m
     compute s12(k) = sum
*
     compute s22(k) = 1-s11(k)-s21(k)-s12(k)
end do k
*
do k = 5,134
     compute s1(k) = s11(k+1) + s21(k+1)
     compute s2(k) = 1 - s1(k)
end do k
compute s1(135) = s11(135) + s12(135)
compute s2(135) = 1 - s1(135)
*
compute sums11 = sums22 = sums1 = sums2 = 0
do k = 6,135
     compute sums11 = sums11 + s11(k)
     compute sums22 = sums22 + s22(k)
     compute sums1 = sums1 + s1(k-1)
     compute sums2 = sums2 + s2(k-1)
end do k
compute p11 = sums11/sums1
compute p22 = sums22/sums2
end trans
********** gradient procedure **********
procedure grad
compute ap = %abs(p)
do j = 1,6
     compute eps(j) = %max(.00001,.00001*ap(j))
end do j
do j = 1,6
*
     compute p(j) = p(j) + eps(j)
     execute pwsr
     compute g1 = wsr
*
     compute p(j) = p(j) - 2*eps(j)
     execute pwsr
     compute g2 = wsr

     compute g(j) = (g1-g2)/(2*eps(j))
     compute p(j) = p(j) + eps(j)
end do j
end grad
********** optimal step procedure **********
procedure opstep
do k = 1,10
     compute s = gs(k)
     compute dp = s*g
     compute p = p + dp
     execute pwsr
     compute vals(k)= wsr
     compute p = lp
end do k
do k = 1,10
   compute chk = %valid(vals(k))
   if chk == 0
        compute vals(k) = -100000000.00
   else
        compute vals(k) = vals(k)
end do k
*
compute max = %maxvalue(vals)
compute j = 0
do k = 1,10
     if vals(k) == max
       compute j = k
     else
       compute j = j
end do k
end opstep
********** s.vals & constant parameters *********
compute p(1) = th(1) = -0.0683 ; compute p(2) = th(2) =  1.20
compute p(3) = th(3) =  0.1419 ; compute p(4) = th(4) =  0.0577
compute p(5) = th(5) = -0.1493 ; compute p(6) = th(6) = -0.1382
compute p11 = th(8) = .75 ; compute p22 = th(9) = .90
compute v = th(7) = .6501
compute itr = 0
*
compute gs =||0.0,0.0001,0.001,0.01,0.03125,.0625,.125,.25,.5,1||
*************** the e.m. algorithm ***************
infobox(action=define) 'the e.m algorithm'
until dthmax < .001
{
compute itr = itr + 1
infobox(action=modify) %string(itr)
compute lth = th
*
execute resids
execute dist
execute rho
execute filter
execute smoother
*
do sa = 1,10
     compute lp = p
     execute grad
     execute opstep
     compute s = gs(j)
     compute dp = s*g
     compute p = p + dp
end do sa
*
execute trans
execute var
compute th(1) = p(1) ; compute th(2) = p(2)
compute th(3) = p(3) ; compute th(4) = p(4)
compute th(5) = p(5) ; compute th(6) = p(6)
compute th(8) = p11  ; compute th(9) = p22
compute th(7) = v
compute dthmax = %maxvalue(%abs(th-lth))
}
end until
infobox(action=remove)
********** final inferences ***********
execute resids
execute dist
execute rho
execute filter
compute sum = 0
do k = 5,135
     compute sum = sum + log(ucd(k))
end do k
compute llval = sum
execute smoother
********** output *********
display '# of iterations =' itr
display 'mu1 =' p(1)
display 'mu2 =' p(2)
display 'b1 =' p(3)
display 'b2 =' p(4)
display 'b3 =' p(5)
display 'b4 =' p(6)
display 'variance =' v
display 'p11=' p11
display 'p22=' p22
display 'log-likelihood =' llval

























