*********************************************
* Program to estimate Hamilton's Markov     *
* switching model using maximum likelihood. *
* 					    *
* Author:  James Watkins July 1997          *
* Revised: Bruce Mizrach Nov. 29, 1997      *
*                                           *
* Cite:                                     *
* Bruce Mizrach and James Watkins           *
* "A Markov Switching Cookbook"             *
* Rutgers University Working Paper #98-17   *
* in Philip Rothman (ed.)                   *
* "Nonlinear Time Series Analysis of        *
* Economics and Financial Data"             *
* Dordrecht: Kluwer, forthcoming.           *
*********************************************
allocate 135
open data c:\rutgers\econ510\problem\hamilton_gdp.dat
data(org=obs,format=free) 1 135 yearq dgdp
* Make the second sub-sample begin with 1 by clipping the old data set
* data(org=obs,format=free) 1 46 yearq dgdp
set y = dgdp
* set nfinish to the last sample point
compute nfinish=135
* convergence tolerance and max iterations
compute tolerance=0.005
compute itermax=10
********** global variable declarations **********
declare vector p(9)
declare vector lp(9)
declare real p11 p22 q1 q2
declare rectangular r(16,135)
declare rectangular b(32,135)
declare rectangular d(32,135)
declare vector ucd(135)
declare real sum llval ll
*
declare vector ap(9) eps(9)
declare vector dp(9)
declare real g1 g2 s
declare rectangular a(9,9)
declare vector g(9) lg(9) dg(9)
declare vector gs(14) vals(14)
declare real max step chk
declare rectangular dgt(1,9) dpt(1,9)
declare rectangular z(1,1)
declare real z1 z2
*
declare integer i k m n w j itr
declare real dpmax
*
declare vector h1(9) h2(9)
declare rectangular h(9,9)
declare vector se(9)
********** distribution functions ***********
frml c1 = 1/((2*%pi*abs(p(7)))**.5)
frml c2 = 1/(2*abs(p(7)))
frml e1 = y(T) - p(1)
frml e2 = y(T) - p(2)
* (01):1,1,1,1,1
frml f01 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (02):2,1,1,1,1
frml f02 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (03):1,2,1,1,1
frml f03 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (04):2,2,1,1,1
frml f04 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (05):1,1,2,1,1
frml f05 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (06):2,1,2,1,1
frml f06 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (07):1,2,2,1,1
frml f07 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (08):2,2,2,1,1
frml f08 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e1(T-4))**2)
* (09):1,1,1,2,1
frml f09 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (10):2,1,1,2,1
frml f10 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (11):1,2,1,2,1
frml f11 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (12):2,2,1,2,1
frml f12 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (13):1,1,2,2,1
frml f13 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (14):2,1,2,2,1
frml f14 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (15):1,2,2,2,1
frml f15 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (16):2,2,2,2,1
frml f16 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e1(T-4))**2)
* (17):1,1,1,1,2
frml f17 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (18):2,1,1,1,2
frml f18 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (19):1,2,1,1,2
frml f19 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (20):2,2,1,1,2
frml f20 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (21):1,1,2,1,2
frml f21 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (22):2,1,2,1,2
frml f22 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (23):1,2,2,1,2
frml f23 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (24):2,2,2,1,2
frml f24 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e1(T-3)-p(6)*e2(T-4))**2)
* (25):1,1,1,2,2
frml f25 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (26):2,1,1,2,2
frml f26 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (27):1,2,1,2,2
frml f27 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (28):2,2,1,2,2
frml f28 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e1(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (29):1,1,2,2,2
frml f29 = c1*exp(-c2*(e1(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (30):2,1,2,2,2
frml f30 = c1*exp(-c2*(e2(T)-p(3)*e1(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (31):1,2,2,2,2
frml f31 = c1*exp(-c2*(e1(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
* (32):2,2,2,2,2
frml f32 = c1*exp(-c2*(e2(T)-p(3)*e2(T-1)-p(4)*e2(T-2)-p(5)*e2(T-3)-p(6)*e2(T-4))**2)
********** probability constraints **********
frml  s1 = exp(p(8))/(1+exp(p(8)))
frml  s2 = exp(p(9))/(1+exp(p(9)))
********** distribution procedure **********
procedure dist
do m = 5,nfinish
   compute d(1,m) = f01(m) ; compute d(2,m) = f02(m)
   compute d(3,m) = f03(m) ; compute d(4,m) = f04(m)
   compute d(5,m) = f05(m) ; compute d(6,m) = f06(m)
   compute d(7,m) = f07(m) ; compute d(8,m) = f08(m)
   compute d(9,m) = f09(m) ; compute d(10,m) = f10(m)
   compute d(11,m) = f11(m) ; compute d(12,m) = f12(m)
   compute d(13,m) = f13(m) ; compute d(14,m) = f14(m)
   compute d(15,m) = f15(m) ; compute d(16,m) = f16(m)
   compute d(17,m) = f17(m) ; compute d(18,m) = f18(m)
   compute d(19,m) = f19(m) ; compute d(20,m) = f20(m)
   compute d(21,m) = f21(m) ; compute d(22,m) = f22(m)
   compute d(23,m) = f23(m) ; compute d(24,m) = f24(m)
   compute d(25,m) = f25(m) ; compute d(26,m) = f26(m)
   compute d(27,m) = f27(m) ; compute d(28,m) = f28(m)
   compute d(29,m) = f29(m) ; compute d(30,m) = f30(m)
   compute d(31,m) = f31(m) ; compute d(32,m) = f32(m)
end do m
end dist
********** rho procedure **********
procedure rho
compute p11 = s1(1) ; compute p22 = s2(1)
compute q1 = (1-p22)/(2-p11-p22)
compute q2 = (1-p11)/(2-p11-p22)
*
compute r(1,4)  = p11*p11*p11*q1
compute r(2,4)  = (1-p11)*p11*p11*q1
compute r(3,4)  = (1-p22)*(1-p11)*p11*q1
compute r(4,4)  = p22*(1-p11)*p11*q1
compute r(5,4)  = p11*(1-p22)*(1-p11)*q1
compute r(6,4)  = (1-p11)*(1-p22)*(1-p11)*q1
compute r(7,4)  = (1-p22)*p22*(1-p11)*q1
compute r(8,4)  = p22*p22*(1-p11)*q1
*
compute r(9,4)  = p11*p11*(1-p22)*q2
compute r(10,4) = (1-p11)*p11*(1-p22)*q2
compute r(11,4) = (1-p22)*(1-p11)*(1-p22)*q2
compute r(12,4) = p22*(1-p11)*(1-p22)*q2
compute r(13,4) = p11*(1-p22)*p22*q2
compute r(14,4) = (1-p11)*(1-p22)*p22*q2
compute r(15,4) = (1-p22)*p22*p22*q2
compute r(16,4) = p22*p22*p22*q2
*
end rho
********** recursive filter procedure **********
procedure filter
compute llval = 0
do n = 5,nfinish
     compute sum = 0
     do m = 1,29,4
          compute w = (m+1)/2
          compute b(m,n)   = d(m,n)*p11*r(w,n-1)
          compute b(m+1,n) = d(m+1,n)*(1-p11)*r(w,n-1)
          compute b(m+2,n) = d(m+2,n)*(1-p22)*r(w+1,n-1)
          compute b(m+3,n) = d(m+3,n)*p22*r(w+1,n-1)
          compute sum = sum + b(m,n) + b(m+1,n) + b(m+2,n) + b(m+3,n)
     end do m
     compute ucd(n) = sum
     compute llval = llval + log(sum)
     do m = 1,16
          compute r(m,n) = (b(m,n) + b(m+16,n))/ucd(n)
     end do m
end do n
end filter
********** gradient procedure **********
procedure grad
compute ap = %abs(p)
do j = i,9
     compute eps(j) = %max(.00001,.00001*ap(j))
end do j
do j = i,9
*
     compute p(j) = p(j) + eps(j)
     execute dist
     execute rho
     execute filter
     compute g1 = llval
*
     compute p(j) = p(j) - 2*eps(j)
     execute dist
     execute rho
     execute filter
     compute g2 = llval

     compute g(j) = (g1-g2)/(2*eps(j))
     compute p(j) = p(j) + eps(j)
end do j
end grad
********** optimal step procedure **********
procedure opstep
do k = 1,14
     compute s = gs(k)
     compute dp = s*a*g
     compute p = p + dp
     execute dist
     execute rho
     execute filter
     compute vals(k)= llval
     compute p = lp
end do k
do k = 1,14
   compute chk = %valid(vals(k))
   if chk == 0
        compute vals(k) = -100000000.00
   else
        compute vals(k) = vals(k)
end do k
*
compute max = %maxvalue(vals)
compute j = 0
do k = 1,14
     if vals(k) == max
       compute j = k
     else
       compute j = j
end do k
end opstep
********** s.vals & constant parameters *********
compute p(1) = -0.0683 ; compute p(2) =  1.20
compute p(3) =  0.1419 ; compute p(4) =  0.0577
compute p(5) = -0.1493 ; compute p(6) = -0.1382
compute p(7) = .6501
compute p(8) = log(.75) - log(1-.75)
compute p(9) = log(.90) - log(1-.90)
*
compute gs =||0.0,0.03125,.0625,.125,.25,.5,1,2,4,8,16,32,64,128||
compute i = 1
*************** 'prealgorithm' ***************
execute dist
execute rho
execute filter
execute grad
*************** optimization via d.f.p. ***************
infobox(action=define) 'd.f.p. '
compute a = .0001*%identity(9)
compute itr = 1; compute iter=1;
*
while itr < itermax
{
compute itr  = itr + 1
compute iter = iter + 1
infobox(action=modify) %string(itr-1)
compute lp = p
compute lg = g
********** optimal step size **********
execute opstep
compute s = gs(j)
********** updated parameter vector **********
compute dp = s*a*g
compute p = p + dp
********** updated gradient **********
execute grad
********** updated inverse hessian **********
compute dg = g - lg
do k = 1,9
     compute dgt(1,k) = dg(k) ; compute dpt(1,k) = dp(k)
end do k
compute z = dgt*a*dg
compute z1 = z(1,1)**(-1)
compute z = dgt*dp
compute z2 = z(1,1)**(-1)
compute a = a - z1*(a*dg*dgt*a) - z2*(dp*dpt)
********** convergence test **********
compute dpmax = %maxvalue(%abs(p-lp))
if dpmax < tolerance  {
	compute itr=itermax
}
}
end while
infobox(action=modify) "Computing standard errors"
********** final inferences ***********
execute dist
execute rho
execute filter
compute ll = llval
********** standard errors **********
compute p(8) = p11
compute p(9) = p22
frml s1 = p(8)
frml s2 = p(9)
*
do i = 1,9
     compute p(i) = p(i) + eps(i)
     execute grad
     compute h1 = g
     compute p(i) = p(i) - 2*eps(i)
     execute grad
     compute h2 = g
     do j = i,9
          compute h(i,j) = (h1(j)-h2(j))/(2*eps(j))
     end do j
     compute p(i) = p(i) + eps(i)
end do i
*
do i = 2,9
     do j = 1,i-1
          compute h(i,j) = h(j,i)
     end do j
end do i
*
compute a = (-1)*h
compute a = inv(a)
do i = 1,9
     compute se(i) = (a(i,i))**(.5)
end do i
infobox(action=remove)
********** output *********
display '# of iterations =' iter
display 'mu1 =' p(1) ; display 's.e. =' se(1)
display 'mu2 =' p(2) ; display 's.e. =' se(2)
display 'b1 =' p(3) ; display 's.e. =' se(3)
display 'b2 =' p(4) ; display 's.e. =' se(4)
display 'b3 =' p(5) ; display 's.e. =' se(5)
display 'b4 =' p(6) ; display 's.e. =' se(6)
display 'variance =' p(7) ; display 's.e. =' se(7)
display 'p11=' p11 ; display 's.e. =' se(8)
display 'p22=' p22 ; display 's.e. =' se(9)
display 'log of likelihood = ' ll










































